from .utils import Formula
from .utils import build_problem_formula, build_Student_AnswersAndScores_for_SingleProblem_from_pth
from .utils import build_problem_formulas_for_multiProblems, build_Student_AnswersAndScores_for_MultiProblem_from_dict
from .utils import multiStudent_compare_multiProblem
from .utils import is_almost_equivalent,comparing_rel, whether_rel_latex_correct, whether_rel_latex_correct_with_units, whether_rel_latex_correct_with_units_with_only_one_dict_parameter
from .utils import compare_multiple_formula_pairs
# note version.
__version__ = '1.0.1'